///////////////////////////////////////////////////////////////////////////////
// Namespace.java - Sets name/value pairs for the Template package.
// Copyright (C) 2003 Michael A. Ericson (maerics@cs.unm.edu)
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details
///////////////////////////////////////////////////////////////////////////////
package com.maerics.template;


import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


class Namespace {
  // Private representation.
  private Hashtable _hashtable;
  private Hashtable _classes;
  private Namespace _parent;
  private boolean _caseInsensitive;
  private boolean _sloppyNaming;
  private boolean _globalVars;
  private boolean _loopContext;


  // How to tell if a method is a JavaBean property.
  private static final Pattern PROPERTY_REGEX = Pattern.compile("^get([A-Z][a-zA-Z0-9]+)$");


  /**
   * Constructs a new, empty, namespace with the specified attributes.
   *
   * @param caseInsensitive true if identifiers should not be case sensitive.
   * @param sloppyNaming true if camel-hump identifiers should be duplicated
   *     in underscore style.
   * @param loopContext true if loop context variables should be provided.
   * @param globalVars true if lookups fallback to more general scopes.
   */
  public Namespace(boolean caseInsensitive,
		   boolean sloppyNaming,
		   boolean globalVars,
		   boolean loopContext)
  {
    // Configuration options.
    this._caseInsensitive = caseInsensitive;
    this._sloppyNaming = sloppyNaming;
    this._globalVars = globalVars;
    this._loopContext = loopContext;

    // The hashtable of name/vals and the hash of class properties.
    this._hashtable = new Hashtable();
    this._classes = new Hashtable();
  }


  /**
   * Clears all values from the namespace.
   */
  public void clear() {
    this._hashtable.clear();
  }


  /**
   * Clears a single value from the namespace.
   */
  public void clear(String name) {
    this._hashtable.remove(name);
  }


  /**
   * Adds a new name/value relationship to the namespace.
   */
  public void define(String name, Object value) {
    // First check the arguments.
    if ( name == null && (value instanceof String || value instanceof Vector) )
      throw new IllegalArgumentException(
	  "Can't define String or Vector with null name");

    // If the value is a vector, then create a vector of namespaces.
    if ( value instanceof Vector ) {
      Vector values = (Vector) value;
      Vector<Namespace> vector = new Vector<Namespace>();
      this._put(name, vector);

      // Iterate through each element and add it's value to the namespace.
      int size = values.size();
      int i = 0;
      for ( Iterator it = values.iterator(); it.hasNext(); i++ ) {

	// Create a new namespace with inherited attributes.
	Namespace n = new Namespace(
	  this._caseInsensitive,
	  this._sloppyNaming,
	  this._globalVars,
	  this._loopContext
	);

	// Each element must be either a map or an object with bean properties.
	Object element = it.next();
	n.define(null, element);
	n.setParent(this);

	// Set the loop context variables, if needed.
	if ( this._loopContext ) {
	  // Make some nice zero padding, first.
	  int totalDigits = new Integer(size).toString().length();
	  StringBuffer buf = new StringBuffer();
	  for ( int j = 0; j < totalDigits; j++ )
	    buf.append("0");
	  String zeroes = buf.toString();

	  // Actually add all the __first__, __index__, etc. vars.
	  Namespace._makeLoopContextVariables(n, i, values.size(), zeroes);
	}
	vector.add(n);
      }

      // Done adding a vector to the namespace.
      return;
    }

    // If the object is a boolean wrapper then set silently or clear.
    if ( value instanceof Boolean ) {
      Boolean b = (Boolean) value;
      if ( b.booleanValue() == true ) {
	this._put(name, "");
      } else {
	this.clear(name);
      }
    }

    // If the object is any kind of primitive wrapper then add it as a string.
    if ( value instanceof Short || value instanceof Integer
	|| value instanceof Long || value instanceof BigInteger
	|| value instanceof Float || value instanceof Double
	|| value instanceof BigDecimal )
    {
      this._put(name, value+"");
    }

    // Add the map of name/value pairs.
    if ( value instanceof Map ) {
      Map map = (Map) value;
      for ( Iterator it = map.keySet().iterator(); it.hasNext(); ) {
	String key = (String) it.next();
	Object val = map.get(key);
	this.define(key, val);
      }
    }

    // Add the plain string value.
    if ( value instanceof String ) {
      this._put(name, (String) value);
      return;
    }

    // Finally, just add all of the object's bean properties.
    this.defineAsBean(name, value);
  }


  /**
   * Adds all of the JavaBean (TM) properties to the namespace.  If the
   * parameter "name" is non-null then it's value is prepended to each Bean
   * property value preceeded by a dot, e.g. <code>user.FirstName</code>.
   */
  private void defineAsBean(String name, Object value) {
    if ( value == null ) return;
    Class c = value.getClass();
    Vector properties = (Vector) this._classes.get(c);

    // If we don't already know the attributes of this class, then store them.
    if ( properties == null ) {
      properties = new Vector();
      this._classes.put(c, properties);

      // Look at all the known methods and query them for properties.
      Method method[] = c.getMethods();
      for ( int i = 0; i < method.length; i++ ) {
	String methodName = method[i].getName();

	// If the method name is a bean property, then store the name.
	Matcher m = PROPERTY_REGEX.matcher(methodName);
	if ( m.matches() && !m.group(1).equals("Class")
	    && method[i].getParameterTypes().length == 0 )
	{
	  String propName = m.group(1);
	  Object methodAndName[] = new Object[] { method[i], propName };
	  properties.add(methodAndName);
	}
      }
    }

    // Iterate through each property and set it's value in the namespace.
    for ( Iterator it = properties.iterator(); it.hasNext(); ) {
      Object methodAndName[] = (Object[]) it.next();
      Method method = (Method) methodAndName[0];
      String methodName = (String) methodAndName[1];

      // Now ask for the property name from the object
      try {
	String varName = (name == null) ? methodName : name+"."+methodName;
//	String varValue = method.invoke(value, null).toString();
	String varValue = method.invoke(value).toString();

	// And define it in the namespace.
	this.define(varName, varValue);

	// TODO: We could really warn about "public" classes better.
      } catch (IllegalAccessException iae) { // ignore...
      } catch (IllegalArgumentException iae) { // ignore...
      } catch (InvocationTargetException ite) { // ignore...
      } catch (NullPointerException npe) { // ignore...
      } catch (ExceptionInInitializerError eie) { // ignore...
      }
    }
  }


  /**
   * Looks up a string value from this namespace.
   */
  public String lookup(String name) {

    // See if we're doing case insensitive naming.
    if ( this._caseInsensitive )
      name = name.toLowerCase();

    // Get the value and see how to return it.
    Object value = this._hashtable.get(name);
//System.err.println("NAMESPACE: Lookup \""+name+"\"=\""+value+"\"");
    if ( value == null ) {
      return (this._globalVars && this._parent!=null)
	? this._parent.lookup(name)
	: null;
    } else if ( value instanceof Vector ) {
      return ((Vector) value).size()+"";
    } else if ( value instanceof String ) {
      return (String) value;
    } else {
      throw new IllegalStateException(
	  "INTERNAL ERROR: Don't know how to lookup an object of type "
	  + value.getClass().getName());
    }
  }


  /**
   * Looks up an iterator value from this namespace.
   */
  public Vector lookupIterator(String name) {
    // See if we're doing case insensitive naming.
    if ( this._caseInsensitive )
      name = name.toLowerCase();

    // Get the value and see how to return it.
    Object value = this._hashtable.get(name);
    if ( value == null || value instanceof String ) {
      return null;
    } else if ( value instanceof Vector ) {
      Vector vector = (Vector) value;
      return (vector.size() > 0) ? vector : null;
    } else {
      throw new IllegalStateException(
	  "INTERNAL ERROR: Don't know how to return an iterator of type "
	  + value.getClass().getName());
    }
  }


  //
  // Does the dirty work of setting a value in the namespace.
  //
  private void _put(String name, Object value) {
    // If we're doing sloppy naming then make the two identifiers.
    if ( this._sloppyNaming ) {
      String n1 = Namespace.id2java(name);
      String n2 = Namespace.java2id(n1);

      // Lower case them, if we don't care about case.
      if ( this._caseInsensitive ) {
	n1 = n1.toLowerCase();
	n2 = n2.toLowerCase();
      }

      this._hashtable.put(n1, value);
//System.err.println("NAMESPACE: "+this+" (SLOP) put "+n1+" = "+value);
      this._hashtable.put(n2, value);
//System.err.println("NAMESPACE: "+this+" (SLOP) put "+n2+" = "+value);
      return;
    }

    // If we don't care about case then make it lower.
    if ( this._caseInsensitive ) {
      this._hashtable.put(name.toLowerCase(), value);
//System.err.println("NAMESPACE: "+this+" (CASE) put "+name.toLowerCase()+" = "+value);
    } else {
      this._hashtable.put(name, value);
//System.err.println("NAMESPACE: "+this+" (STD) put "+name+" = "+value);
    }
  }


  //
  // Sets the parent of this namespace
  //
  private void setParent(Namespace ns) {
    this._parent = ns;
  }


  //
  // Returns a hash of loopContextVar=>value for an iteration.
  //
  private static void _makeLoopContextVariables(Namespace ns,
						int i,
						int size,
						String padding)
  {
    // Set the loop indices.
    ns._hashtable.put("__index__", (i+1)+"");
    ns._hashtable.put("__zindex__", i+"");
    String zp = new String(padding + i);
    ns._hashtable.put("__pindex__",
	zp.substring(zp.length()-padding.length(), zp.length()));

    // Set a true value if it's the first element of the loop.
    if ( i == 0 ) {
      ns._hashtable.put("__first__", "");
      ns._hashtable.put("__outer__", "");
    }

    // Set a true value if it's the last element of the loop.
    if ( i == size-1 ) {
      ns._hashtable.put("__last__", "");
      ns._hashtable.put("__outer__", "");
    }

    // Set a true value if it's neither the first nor last.
    if ( i != 0 && i != size-1 ) {
      ns._hashtable.put("__inner__", "");
    }

    // Set a true value stating if the index is even or odd.
    if ( (i+1) % 2 == 0 ) {
      ns._hashtable.put("__even__", "");
    } else {
      ns._hashtable.put("__odd__", "");
    }
  }


  /**
   * Returns the string identifier in Java "camel-hump" style.  This method
   * is just a wrapper for <code>Namespace.id2java([string], false)</code>.
   *
   * @see #id2java(String,boolean)
   */
  private static String id2java(String s) { return id2java(s, true); }


  /**
   * Returns the given string identifier in Java "camel-hump" style.
   * Conversion is only attempted if the identifier contains spaces, dashes,
   * or underscores.  Here are some examples showing the input, the output in
   * global context and the output in local context:
   * <pre>
   *    INPUT			GLOBAL		    LOCAL
   *    number_of_cats       -- NumberOfCats     -- numberOfCats
   *    _HAVE__A__TRY_       -- HaveATry         -- haveATry
   *    scheme-is-great      -- SchemeIsGreat    -- schemeIsGreat
   *    this_ is _great      -- ThisIsGreat      -- thisIsGreat
   *    AlmostThere          -- AlmostThere      -- almostThere 
   * </pre>
   *
   * @param s the string to format like a java identifier.
   * @param globalName true if this identifier is in global scope and should
   *     start with a lower case letter, false if it should start uppercase.
   * @return a string suitable for use as a variable name in java, the empty
   *     string if the input is null.
   */
  private static String id2java(String s, boolean globalName) {
    if ( s == null || s.equals("") ) return "";
    StringBuffer javaname = new StringBuffer();

    // Only process the identifier if there are spaces or underscores.
    if ( s.indexOf(' ')>=0 || s.indexOf('_')>=0 || s.indexOf('-')>=0 ) {
      boolean nextIsCap = true;

      // For each character in the given string...
      for ( int i = 0; i < s.length(); i++ ) {
	// I chose String.charAt() rather than String.toCharArray().
	String ch = new Character(s.charAt(i)).toString();

	if ( ch.equals("_") || ch.equals(" ") || ch.equals("-") ) {
	  nextIsCap = true;
	} else if ( nextIsCap ) {
	  javaname.append(ch.toUpperCase());
	  nextIsCap = false;
	} else {
	  javaname.append(ch.toLowerCase());
	}
      }
    } else {
      javaname.append(s);
    }

    // Now fix the first-character capitalization.
    if ( globalName ) {
      String firstChar = new Character(javaname.charAt(0)).toString();
      javaname.setCharAt(0, firstChar.toUpperCase().charAt(0));
    } else {
      String firstChar = new Character(javaname.charAt(0)).toString();
      javaname.setCharAt(0, firstChar.toLowerCase().charAt(0));
    }
    return javaname.toString();
  }


  /**
   * Returns the given Java "camel" hump identifier in underscore style.
   */
  public static String java2id(String s) {
    if ( s == null || s.equals("") ) return "";
    StringBuffer name = new StringBuffer();

    // Skip ahead to the first character after the dot.
    int firstChar = 0;
    int firstDotIndex = s.indexOf(".");
    if ( firstDotIndex > -1 ) {
      name.append(s.substring(0, firstDotIndex+1));
      firstChar = firstDotIndex + 1;
    }

    // Now do the name conversion.
    for ( int i = firstChar; i < s.length(); i++ ) {
      char ch = s.charAt(i);
      if ( Character.isUpperCase(ch) && i != firstChar )
	name.append("_");
      name.append(new Character(ch).toString().toLowerCase());
    }
    return name.toString().toLowerCase();
  }
}

