///////////////////////////////////////////////////////////////////////////////
// Template.java - A text file template with variables, conditionals and loops.
// Copyright (C) 2003 Michael A. Ericson (maerics@cs.unm.edu)
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
///////////////////////////////////////////////////////////////////////////////
package com.maerics.template;


import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * A text file template class that supports variable substitutions, conditional
 * statements, and iteration (looping). Text file templates are useful for
 * separating content from presentation, an essential practice when employing
 * the Model-View-Controller (MVC) paradigm.  This template class reads a
 * text template and parses out markup tags that specify commands to modify the
 * template; such as, substituting variables, executing conditional statements
 * (if/then/else), and others.  This class is truly a generic <i>text</i>
 * template class in that it presupposes nothing about the structure of the
 * template file itself, meaning it can be used for plain-text files, HTML
 * files, XML files, or any other text based file.  This means, in particular,
 * that tags cannot be "hidden in comments" (such as in an HTML file) to be
 * ignored.
 *
 * <h3>Template Markup Syntax</h3>
 *
 * <p>The <code>Template</code> class recognizes the following SGML'ish
 * markup tags:
 * <ul>
 *   <li><code>&lt;tmpl_var <i>name</i>&gt;</code>
 *   <li><code>&lt;tmpl_if <i>cond</i>&gt;...&lt;/tmpl_if&gt;</code><br>
 *   <li><code>&lt;tmpl_unless <i>cond</i>&gt;...&lt;/tmpl_unless&gt;</code><br>
 *   <li><code>&lt;tmpl_else&gt;</code><br>
 *   <li><code>&lt;tmpl_loop <i>name</i>&gt;...&lt;/tmpl_loop&gt;</code>
 * </ul>
 * These tags must be properly nested.</p>
 *
 * <h3>Template Variables</h3>
 *
 * <p>Most template tags (var, if, unless, loop) require a single argument
 * which is either the name of a variable, a conditional, or an iterator.  Most
 * of these values can be used interchangeably but with different effects.  For
 * example, <code>&lt;tmpl_var <i>name</i>&gt;</code> tags require a single
 * argument which is treated as a plain variable.  When the template output is
 * requested the entire tag will be replaced with the value to which "name" was
 * set.  For example, if you call <code>myTemplate.set("foo","bar")</code>
 * then all template variable tags whose argument is "foo" will be replaced
 * with the text "bar".
 *
 * <h3>Template Conditionals</h3>
 * 
 * <p>Conditionals consist of a start-tag and an end-tag which define a section
 * of text that will only be included in the output if the named variable is
 * "true."  A variable is considered to be true if it has any value set for
 * it, that is, if any form of <code>Template.set(name, value)</code> has been
 * called for the given conditional name where the value is not null or an
 * empty iterator (one with no elements).  Note that if a variable is set to
 * the empty string (<code>""</code>) it will evaluate to "true" as a
 * conditional but will not display any text when used as a variable since
 * its value is an empty string.  Conditionals can also take an "else"
 * block that will be included in the output if the named condition does not
 * evaluate to true.</p>
 *
 * <h3>Template Iterators</h3>
 *
 * <p>Iterators are implicit conditionals in that if the named iterator does
 * not exist then the enclosed section of text is not included in the output.
 * Similarly, they can contain an "else" clause.  If an iterator's name does
 * resolve to "true" and it is a vector of objects, then each object will be
 * converted into it's own namespace (either as a "Map" or by inspecting the
 * object's bean properties).  The text enclosed by the "loop" tag
 * and it's corresponding closing tag (or "else" tag) creates a block of text.
 * This block is evaluated in the namespace of each map in the vector, thus
 * "looping" over the block for each namespace provided.  If an iterator
 * is used in a "variable" context the value is the number of elements in the
 * iterator.</p>
 *
 * <h3>Loop Context Variables</h3>
 *
 * <p>If a template is constructed with the attribute
 * "<code>loop_context_variables</code>" set to "true" then the following
 * variables will be automatically provided to each iteration of the loop:</p>
 *
 * <p><ul>
 *   <li><code>__index__</code>: the one-based index number of this iteration.
 *   <li><code>__zindex__</code>: the zero-based index number of this
 *     iteration.
 *   <li><code>__pindex__</code>: the zero-based, zero-padded index number of
 *     this iteration.
 *   <li><code>__first__</code>: true if this iteration is the first in the
 *     loop.
 *   <li><code>__last__</code>: true if this iteration is the last in the
 *     loop.
 *   <li><code>__inner__</code>: true if this iteration is <i>not</i> the
 *     first or last of the loop.
 *   <li><code>__outer__</code>: true if this iteration <i>is</i> the first or
 *     last of the loop.
 *   <li><code>__even__</code>: true if this iteration's index (one-based) is
 *     an even number.
 *   <li><code>__odd__</code>: true if this iteration's index (one-based) is
 *     an odd number.
 * </ul></p>
 *
 * <p>Note that each of these identifiers is a word surrounded on both sides by
 * two underscores.</p>
 *
 * @author Michael A. Ericson <a href="mailto:maerics@cs.unm.edu">maerics@cs.unm.edu</a>
 * @version 0.16
 */
public class Template {
  // The regular expression for finding markup tags.
  private static final String REGEX = ".*?(<(/)?tmpl_(var|if|else|unless|loop)(\\s+([A-Za-z0-9_.-]+))?\\s*>).*?";
  private static final Pattern PATTERN = Pattern.compile(REGEX, Pattern.DOTALL | Pattern.CASE_INSENSITIVE);

  // Contstants that serve as regex grouping placeholders.
  private static final int M_ENTIRE_TAG = 1;
  private static final int M_CLOSING_TAG = 2;
  private static final int M_DIRECTIVE = 3;
  private static final int M_VARIABLE = 5;

  // The private representation.
  private Namespace _namespace;
  private String _template;
  private String _output;
  private Vector _pieces;
  private boolean _hasChanged;

  // Constructor stuff.
  private String _filename = null;
  private boolean _caseInsensitive = false;
  private boolean _sloppyNaming = false;
  private boolean _globalVariables = false;
  private boolean _loopContextVariables = false;


  /**
   * Creates a new Template object with a map of attributes defining the new
   * template's behavior. The parameter to the constructor is a
   * <code>Map</code> that contains a number of attributes.  These attributes
   * determine how the newly created template object will resolve and
   * substitute variable names. This parameter must contain the required
   * attributes, listed below.
   *
   * <h4>Required Attributes:</h4>
   *
   * <p><ul>
   *   <li><code>filename</code> - The string name of the file that contains
   *     the template.<br>
   *   &nbsp; &nbsp; <i>OR</i><br>
   *   <li><code>text</code> - The text string of the template itself (will be
   *     duplicated).
   * </ul></p>
   *
   * <h4>Optional Attributes:</h4>
   *
   * <p><ul>
   *   <li><code>case_insensitive</code>: type <code>Boolean</code>
   *     - default <code>Boolean.FALSE</code>.<br>
   *     - Whether or not variable names should be case insensitive.
   *   <li><code>sloppy_naming</code>: type <code>Boolean</code>
   *     - default <code>Boolean.FALSE</code>.<br>
   *     - Allow for automatic conversion from camel-hump to underscore
   *       naming styles and vice-versa.  For example the variable name
   *       "<code>FirstName</code>" would be automatically aliased as the name
   *       "<code>first_name</code>" and <i>vice versa</i>.
   *   <li><code>globals_variables</code>: type <code>Boolean</code>
   *     - default <code>Boolean.FALSE</code>.<br>
   *     - Variables can be accessed from any sub-scope of the namespace if
   *       they are not shadowed.
   *   <li><code>loop_context_variables</code>: type <code>Boolean</code>
   *     - default <code>Boolean.FALSE</code>.<br>
   *     - Allow for loop context variables like <code>__index__</code>,
   *       <code>__first__</code>, etc.
   * </ul></p>
   *
   * <p>The required arguments, as described, must be either a path to a text
   * file that contains the template, or the text of the template itself. The
   * optional arguments are boolean values denoted by the strings
   * "<code>true</code>" or "<code>false</code>".</p>
   *
   * <h4>Example</h4>
   *
   * A template might be constructed as such:
   * <p><pre>
   *     Hashtable attr = new Hashtable();
   *     attr.put("filename", "./path/to/my/template");
   *     attr.put("case_insensitive", Boolean.TRUE);
   *     attr.put("global_variables", Boolean.TRUE);
   *     attr.put("loop_context_variables", Boolean.TRUE);
   *
   *     Template tmpl = null;
   *     try {
   *         tmpl = new Template(attr);
   *     } catch (FileNotFoundException fnfe) {
   *         System.err.println("ERROR: File not found: "+fnfe.getMessage());
   *         System.exit(1);
   *     } catch (IOException ioe) {
   *         System.err.println("ERROR: I/O exception: "+ioe.getMessage());
   *         System.exit(1);
   *     } catch (TemplateException te) {
   *         System.err.println(e.getMessage());
   *         System.exit(1);
   *     }</pre></p>
   *
   * <p>A little verbose, yes, but flexible. Of course, you'll probably 
   * want to write a method that encapsulates this functionality so you
   * don't have to write it over and over again.
   *
   * @param attributes a map of attributes defining the behavior of the
   *     template.
   *
   * @throws IllegalArgumentException if there is no "filename" or "text"
   *     attribute in the supplied map.
   * @throws FileNotFoundException if the file specified by the
   *     "<code>filename</code>" attribute could not be found.
   * @throws IOException if there was any error opening or reading the file
   *     specified by the "<code>filename</code>" attribute.
   * @throws TemplateException if there was any error parsing the given
   *     template text or constructing the template object.
   */
  public Template(Map attributes) throws FileNotFoundException,
   					 IOException,
					 TemplateException
  {
    if ( attributes == null )
      throw new IllegalArgumentException("Attributes can't be null");
    this._init(attributes);
  }


  /**
   * Creates a new template from the given filename with all optional
   * attributes enabled or disabled.
   *
   * @param filename the file that contains the text of the desired template.
   * @param options true if all optional attributes should be included or
   *     false if none should be.
   *
   * @see #Template(Map)
   */
  public Template(String filename, boolean options)
      throws FileNotFoundException,
	     IOException,
	     TemplateException
  {
      System.out.println("Debug filename:" + filename);
    if ( filename == null )
      throw new IllegalArgumentException("Filename can't be null");

    // Just create our own hash and call the Map constructor.
    Hashtable attributes = new Hashtable();
    attributes.put("filename", filename);
    Boolean opts = new Boolean(options);
    attributes.put("case_insensitive", opts);
    attributes.put("sloppy_naming", opts);
    attributes.put("globals_variables", opts);
    attributes.put("loop_context_variables", opts);
    this._init(attributes);
  }


  /**
   * Creates a new template from the given filename with the default
   * attributes.  All attributes other than the filename assume their default
   * values.
   *
   * @param filename the file that contains the text of the desired template.
   *
   * @see #Template(Map)
   */
  public Template(String filename) throws FileNotFoundException,
  					  IOException,
					  TemplateException
  {
    if ( filename == null )
      throw new IllegalArgumentException("Filename can't be null");

    // Just create our own hash and call the Map constructor.
    Hashtable attributes = new Hashtable();
    attributes.put("filename", filename);
    this._init(attributes);
  }


  /**
   * Returns the text string from which this template object was constructed.
   *
   * @return the text that was used to construct this template object.
   */
  public String getText() {
    return this._template;
  }


  /**
   * Returns the name of the file from which this template was constructed.
   * If this object was constructed, for example, by setting the
   * "<code>text</code>" attribute in its constructor argument then no filename
   * exists and null is returned.
   *
   * @see #Template(Map)
   *
   * @return the name of the file from which this template was constructed or
   *     null there was none.
   */
  public String getFilename() {
    return this._filename;
  }


  /**
   * Returns the output of this template.  The output is a text string in which
   * all variable values and iterators have been substituted appropriately.
   *
   * @return the text that results from executing all tag directives in this
   *     template.
   */
  public String output() {

    // First, see if we've changed anything int the template.
    if ( this._hasChanged ) {
      StringBuffer str = new StringBuffer();
      for ( Iterator it = this._pieces.iterator(); it.hasNext(); ) {
	Piece piece = (Piece) it.next();
	str.append(piece.evaluate(this._namespace));
      }
      this._output = str.toString();
      this._hasChanged = false;
    }

    // Return the output.
    return this._output;
  }


  /**
   * Prints this template's output to the given PrintWriter.  This method
   * simply calls <code>out.print(this.output())</code>.
   */
  public void printTo(PrintWriter out) {
    out.print(this.output());
  }


  /**
   * Removes the value for the given name.  This is the easiest way to remove
   * or "unset" a value for a named variable.  You can also call the Object
   * "<code>set</code>" method with a null argument to achieve the same effect.
   *
   * @see #set(String,Object)
   */
  public void clear(String name) {
    this._namespace.clear(name);
    this._hasChanged = true;
  }



  /**
   * Removes all values from this template.  This is the easiest way to clear
   * out the entire template.
   */
  public void clear() {
    this._namespace.clear();
    this._hasChanged = true;
  }


  /**
   * Sets a boolean value for the named variable.  If the value is "true" then
   * the variable tag will be substituted with the empty string; if the value
   * is false then the variable name will be cleared.
   *
   * @param name the name of the variable to set as true or false.
   * @param value whether or not this variable should make conditionals
   *     succeed.
   */
  public void set(String name, boolean value) {
    if ( value ) {
      this.set(name, "");
    } else {
      this.clear(name);
    }
  }


  /**
   * Sets an integer value for the named variable.  Don't forget that the
   * value zero is a "true" value.
   *
   * @param name the name of the variable that will be substituted for an
   *     integer value.
   * @param value the integer value that this variable should assume.
   */
  public void set(String name, int value) {
    this.set(name, value+"");
  }


  /**
   * Sets a decimal value for the named variable.  Don't forget that the
   * value zero is a "true" value.
   *
   * @param name the name of the variable that will be substituted for an
   *     decimal value.
   * @param value the decimal value that this variable should assume.
   */
  public void set(String name, double value) {
    this.set(name, value+"");
  }


  /**
   * Sets the value for the named variable.
   * <ul>
   *   <li>If the value is <code>null</code> then no value for the name is
   *       set -- this is the equivalent of calling <code>clear(name)</code>.</li>
   *   <li>If the object argument is a <code>Vector</code> then it will be
   *       defined as an iterator.</li>
   *   <li>If the object argument is a <code>Map</code> then all of the
   *       key/value pairs will be defined as in the following example.
   *       Consider the following code snippet:
   *       <pre>
   *         Hashtable purchaser = new Hashtable();
   *         purchaser.put("FirstName", "Michael");
   *         purchaser.put("LastName", "Ericson");
   *         myTemplate.set("client", purchaser);
   *       </pre>
   *       As applied to the following template.
   *       <pre>
   *         Hello, &lt;tmpl_var client.FirstName&gt; &lt;tmpl_var client.LastName&gt;.
   *       </pre>
   *       The following output will result: "<code>Hello, Michael Ericson.</code>"</li>
   *   <li>If the object argument is a <code>Boolean</code> then its behavior
   *       will be exactly the same as the primitive <code>boolean</code>
   *       equivalent.</li>
   *   <li>If the object argument is a <code>String</code> then its string
   *       representation will be used.</li>
   *   <li>Otherwise, the object will be inspected for its JavaBean properties
   *       that will then be treated as a map of values.</li>
   * </ul>
   *
   * @param name the name of the variable that will assume a value.
   * @param value the new value for the name; or clears the name if null.
   *
   * @see Template#set(Object)
   */
  public void set(String name, Object value) {
    // Make sure the name isn't null.
    if ( name == null )
      throw new IllegalArgumentException("Can't set variable for null name");

    // See if we are setting or clearing a value.
    if ( value == null ) {
      this.clear(name);
    } else {
      this._namespace.define(name, value);
    }

    this._hasChanged = true;
  }


  /**
   * <p>Sets values for all of the Map or JavaBean<sup><small>TM</small></sup>
   * properties for this object.  If the object argument is a <code>Map</code>
   * then all of the key/value pairs are defined as name/value pairs in the
   * template.  Otherwise, this method uses reflection to determine all
   * of the bean properties for this object and sets them in the template as
   * though they were Map values.  If the class has no bean properties then
   * no action is taken.</p>
   *
   * <h4>Caveats</h4>
   *
   * <p>Note that if a class is not declared as "public" in its class
   * definition line then this method will not detect its bean properties.</p>
   *
   * @param value the object that will be queried for JavaBean properties.
   */
  public void set(Object value) {
    this._namespace.define(null, value);
    this._hasChanged = true;
  }


  /////////////////////////////////////////////////////////////////////////////
  // Initialization and parsing methods. //////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  //
  // Does the actual work of initializing a new Template object.
  //
  private void _init(Map attributes) throws FileNotFoundException,
  					    IOException,
					    TemplateException
  {
    // See if we've got a raw template.
    String templateText = (String) attributes.get("text");
    if ( templateText != null )
      this._template = new String(templateText);

    // If there's no raw template, then read the text from a file.
    if ( this._template == null ) {
      this._filename = (String) attributes.get("filename");

      // Make sure we've got a filename.
      if ( this._filename == null ) {
	throw new IllegalArgumentException(
	    "Must have \"filename\" or \"text\" " +
	    "attribute to parse a template"
	);
      }

      // Read in the entire file as a single string.
      // IDEから実行した場合は，文字化けしないが，独立したアプリとして実行すると文字が化ける。
      // おそらく，UTF-8をSJISとして読んでいる様子。
      // そこで，以下のように，UTF-8と指定してデータを読み込ませる。     
      java.io.FileInputStream fis = new java.io.FileInputStream(this._filename);
      java.io.InputStreamReader isr = new java.io.InputStreamReader(fis,"UTF-8"); 
      java.io.BufferedReader in = new java.io.BufferedReader(isr);     
//    BufferedReader in = new BufferedReader(new FileReader(this._filename));
      System.out.println("Debug this._filename:" + this._filename);
   // java.io.InputStream  resource   = this.getClass().getResourceAsStream(this._filename);
   // System.out.println("Debug resource:" + resource.toString());

   // java.io.InputStreamReader  isr  = new java.io.InputStreamReader(resource);
   // java.io.BufferedReader     fin  = new java.io.BufferedReader(isr);

      
      
      StringBuffer rawTemplateBuffer = new StringBuffer();
      String line;
      while ( (line=in.readLine()) != null ) {
	rawTemplateBuffer.append(line);
	rawTemplateBuffer.append("\n");
      }
      this._template = rawTemplateBuffer.toString();
    }

    // Check the optional parameters and set the attributes accordingly.
    this._caseInsensitive = false;
    this._sloppyNaming = false;
    this._globalVariables = false;
    this._loopContextVariables = false;

    try {
      // See if the user is setting the "case_insensitive" attribute.
      Boolean cases = (Boolean) attributes.get("case_insensitive");
      if ( cases != null && cases.equals(Boolean.TRUE) )
	this._caseInsensitive = true;
    } catch (ClassCastException cce) {
      throw new IllegalArgumentException(
	  "Argument to \"case_insensitive\" must be a Boolean");
    }

    try {
      // See if the user is setting the "sloppy_naming" attribute.
      Boolean slop = (Boolean) attributes.get("sloppy_naming");
      if ( slop != null && slop.equals(Boolean.TRUE) )
	this._sloppyNaming = true;
    } catch (ClassCastException cce) {
      throw new IllegalArgumentException(
	  "Argument to \"sloppy_naming\" must be a Boolean");
    }

    try {
      // See if the user is setting the "global_variables" attribute.
      Boolean global = (Boolean) attributes.get("global_variables");
      if ( global != null && global.equals(Boolean.TRUE) )
	this._globalVariables = true;
    } catch (ClassCastException cce) {
      throw new IllegalArgumentException(
	  "Argument to \"global_variables\" must be a Boolean");
    }

    try {
      // See if the user is setting the "loop_context_variables" attribute.
      Boolean loops = (Boolean) attributes.get("loop_context_variables");
      if ( loops != null && loops.equals(Boolean.TRUE) )
	this._loopContextVariables = true;
    } catch (ClassCastException cce) {
      throw new IllegalArgumentException(
	  "Argument to \"loop_context_variables\" must be a Boolean");
    }

    this._namespace = new Namespace(
	this._caseInsensitive,
	this._sloppyNaming,
	this._globalVariables,
	this._loopContextVariables
    );

    // Finally, parse and construct the template.
    this._pieces = this._parseTemplate(this._template);
    this._hasChanged = true;
  }


  //
  // Does the dirty work of parsing the template text and creating the object.
  // Who needs recursive descent parsing when you've got regular expressions?
  //
  private Vector _parseTemplate(String text) throws TemplateException {
    int charNumber = 1;
    int lineNumber = 1;
    Stack locations = new Stack();

    // Keep track of the text filler and the context.
    StringBuffer textBuffer = new StringBuffer();
    Stack branchNumbers = new Stack();
    Stack branches = new Stack();
    Stack contextVars = new Stack();
    Stack context = new Stack();
    Branch currentBranch;

    // Create the initial branch for the whole template.
    boolean inSecondBranch = false;
    branches.push(currentBranch = new Branch());

    // While we see any directives to parse out of the text.
    for ( Matcher m = PATTERN.matcher(text); m.matches(); ) {
      // Keep track of the current line and character number.
      String substr = text.substring(0, m.start(M_ENTIRE_TAG));
      int numberOfNewlines = countNewlines(substr);
      lineNumber += numberOfNewlines;
      charNumber = ( numberOfNewlines > 0 )
	? m.start(M_ENTIRE_TAG) - substr.lastIndexOf('\n')
	: charNumber + substr.length();

      // Append the plain text piece to the current branch.
      currentBranch.addPiece(new Template.Text(substr));

      // Get the specific attributes of the match.
      String entireTag = m.group(M_ENTIRE_TAG);
      boolean closingTag = (m.group(M_CLOSING_TAG)!=null);
      String directive = m.group(M_DIRECTIVE).toLowerCase();
      String variable = m.group(M_VARIABLE);

      // Process the tag.
      if ( closingTag ) {
	// Make sure they aren't closing a var or else tag.
	if ( directive.equals("var") || directive.equals("else") )
	  throw new TemplateException("Invalid tag "+entireTag, this._filename, lineNumber, charNumber);

	// Make sure we're closing from the right context and add the piece.
	if ( context.size()<1 || !context.peek().equals(directive) ) {
	  throw new TemplateException("Mismatched "+entireTag, this._filename, lineNumber, charNumber);
	} else {

	  // Get the current context and branches off the stack.
	  locations.pop();
	  inSecondBranch = ((Boolean) branchNumbers.pop()).booleanValue();
	  String currentContext = (String) context.pop();
	  String contextVar = (String) contextVars.pop();
	  Branch b1 = (Branch) branches.pop();
	  Branch b2;
	  if ( inSecondBranch ) {
	    b2 = (Branch) branches.pop();
	  } else {
	    b2 = new Branch();
	  }

	  // Create the appropriate piece and add it to the current branch.
	  Piece thePiece = null;
	  if ( currentContext.equals("if") ) { // And "if" clause.
	    thePiece = (inSecondBranch)
	      ? new Template.Conditional(contextVar, true, b2, b1)
	      : new Template.Conditional(contextVar, true, b1, b2);
	  } else if ( currentContext.equals("unless") ) { // An "unless" clause
	    thePiece = (inSecondBranch)
	      ? new Template.Conditional(contextVar, false, b2, b1)
	      : new Template.Conditional(contextVar, false, b1, b2);
	  } else if ( currentContext.equals("loop") ) { // A "loop" clause.
	    thePiece = (inSecondBranch)
	      ? new Template.Each(contextVar, b2, b1)
	      : new Template.Each(contextVar, b1, b2);
	  } else {
	    throw new RuntimeException("INTERNAL ERROR: Invalid closing state \""+currentContext+"\"");
	  }
	  currentBranch = (Branch) branches.peek();
	  currentBranch.addPiece(thePiece);
	  inSecondBranch = false;
	}

      } else if ( directive.equals("else") ) { // Handle an "else" directive.
	// Make sure we are in a conditional context.
	if ( context.size() < 1 || inSecondBranch
	       || (!context.peek().equals("if")
		&& !context.peek().equals("unless")
		&& !context.peek().equals("loop")) ) {
	  throw new TemplateException("Found "+entireTag+" in bad context", this._filename, lineNumber, charNumber);
	} else {
	  // If we're OK, then switch to the second branch.
	  branches.push(currentBranch = new Branch());
	  branchNumbers.push(new Boolean(true));
	  inSecondBranch = true;
	}
      } else if ( directive.equals("var") ) { // Handle a "var" directive.
	// Add the variable template piece to the current branch.
	currentBranch.addPiece(new Template.Variable(variable));
      } else if ( directive.equals("if") 
		  || directive.equals("unless")
		  || directive.equals("loop") ) { // Handle a "conditional".
	// First, make sure we've got a name to go with the conditional.
	if ( variable == null )
	  throw new TemplateException("Found \""+m.group(M_ENTIRE_TAG)+"\" without name", this._filename, lineNumber, charNumber);

	// Put ourselves in the specified context.
	locations.push(new Object[]{new Integer(lineNumber),new Integer(charNumber)});
	branches.push(currentBranch = new Branch());
	branchNumbers.push(new Boolean(inSecondBranch));
	contextVars.push(variable);
	context.push(directive);
	inSecondBranch = false;
      } else {
	throw new TemplateException("INTERNAL ERROR: Invalid directive \""+directive+"\"", this._filename, lineNumber, charNumber);
      }

      // Update the current character number to include the processed tag.
      charNumber += (m.end(M_ENTIRE_TAG) - m.start(M_ENTIRE_TAG));

      // Move on past the matched tag.
      text = text.substring(m.end(M_ENTIRE_TAG), text.length());
      m = PATTERN.matcher(text);
    }

    // Add any remaining text in the string to the template.
    currentBranch.addPiece(new Template.Text(text));

    // When all is said and done, there must be only one piece on the stack.
    if ( branches.size() < 1 ) {
      throw new TemplateException("INTERNAL ERROR: Template out of branches. Please contact Michael Ericson (maerics@cs.unm.edu).", null, -1, -1);
    } else if ( branches.size() > 1 ) {
      String ctx = (String) context.pop();
      Object n[] = (Object[]) locations.pop();
      int ln = ((Integer) n[0]).intValue();
      int cn = ((Integer) n[1]).intValue();
      throw new TemplateException("Unclosed <tmpl_"+ctx+"> at EOF possibly from tag", this._filename, ln, cn);
    }

    // Return the final piece as a vector for our template to iterate over.
    Branch finalBranch = (Branch) branches.pop();
    return finalBranch.getPieces();
  }


  //
  // Returns the number of newline characters ('\n') in the given string.
  //
  private static int countNewlines(String s) {
    int count = 0;
    int offset = 0;

    while ( (offset=s.indexOf('\n',offset)) != -1 ) { count++; offset++; }
    return count;
  }


  /////////////////////////////////////////////////////////////////////////////
  // Template private classes from here on out. ///////////////////////////////
  /////////////////////////////////////////////////////////////////////////////


  //
  // The interface that defines all of the components of a template.
  //
  private static interface Piece { String evaluate(Namespace namespace); }


  //
  // Text pieces just return the text they were given.
  //
  private static class Text implements Piece {
    private String _text;
    public Text(String text) { this._text = text; }
    public String evaluate(Namespace namespace) { return _text; }
  }


  //
  // Variables return the variable definition, if it is defined in the ns.
  //
  private static class Variable implements Piece {
    private String _name;
    public Variable(String name) { this._name = name; }
    public String evaluate(Namespace namespace) {
      String name = namespace.lookup(this._name);
      return (name == null) ? "" : name;
    }
  }


  //
  // A branch is a Piece that can hold other pieces (as a vector).
  //
  private static class Branch implements Piece {
    private Vector _pieces;

    public Branch() { this._pieces = new Vector(); }

    public void addPiece(Piece p) { this._pieces.add(p); }

    public Vector getPieces() { return this._pieces; }

    public String evaluate(Namespace namespace) {
      StringBuffer buffer = new StringBuffer();
      for ( Iterator it = this._pieces.iterator(); it.hasNext(); ) {
	Piece p = (Piece) it.next();
	buffer.append(p.evaluate(namespace));
      }
      return buffer.toString();
    }
  }


  //
  // Conditionals show different branches depending on a variable name lookup.
  //
  private static class Conditional implements Piece {
    private String _name;
    private boolean _positive; // True for "if", false for "unless".
    private Branch _succ;
    private Branch _fail;

    public Conditional(String name, boolean type, Branch succ, Branch fail) {
      if ( name==null ) throw new IllegalArgumentException("INTERNAL ERROR: Can't build a conditional with a null name");
      if ( succ==null ) throw new IllegalArgumentException("INTERNAL ERROR: Can't build a conditional with a null success clause");
      if ( fail==null ) throw new IllegalArgumentException("INTERNAL ERROR: Can't build a conditional with a null failure clause");
      this._name = name;
      this._positive = type;
      this._succ = succ;
      this._fail = fail;
    }

    public Branch getBranch(boolean successful) {
      return (successful)
	? this._succ
	: this._fail;
    }

    public String evaluate(Namespace namespace) {
      String name = namespace.lookup(this._name);

      return ( this._positive && name==null || !this._positive && name!=null)
	? this._fail.evaluate(namespace)
	: this._succ.evaluate(namespace);
    }
  }


  //
  // Each's evaluate a branch across an iterator or evaluate a "false" branch.
  //
  private static class Each implements Piece {
    private String _name;
    private Branch _succ;
    private Branch _fail;

    public Each(String name, Branch succ, Branch fail) {
      this._name = name;
      this._succ = succ;
      this._fail = fail;
    }

    public String evaluate(Namespace namespace) {
      Vector loop = namespace.lookupIterator(this._name);
      StringBuffer ret = new StringBuffer();

      // Add the text from each success piece.
      if ( loop != null ) {
	for ( Iterator it = loop.iterator(); it.hasNext(); ) {
	  // Evaluate the branch for each namespace.
	  Namespace ns = (Namespace) it.next();
	  ret.append(this._succ.evaluate(ns));
	}
      } else {
	// Add the text form the failure piece.
	ret.append(this._fail.evaluate(namespace));
      }

      return ret.toString();
    }
  }
}

